/***************************************************************************
 *   Copyright (C) 2005 by Pavel Vymetálek   *
 *   pvymetalek@seznam.cz   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   --------------------------------------------------------------------  *
 *   SIMEPROM - control program for Elnec's SIMEPROM-01B                   *
 *   Device manufacturer www.elnec.sk                                      *
 *   Linux control program for Linux and other unices by Pavel Vymetálek          *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <simeprom7.h>
#include <hex2bin.h>

extern	boolean Enable_Checksum_Error;
extern	boolean Status_Checksum_Error;

/* We will assume that when one type of addressing is selected, it will be valid for all the
current file. Records for the other type will be ignored. */
extern	unsigned int	Seg_Lin_Select;

extern boolean Starting_Address_Setted;



int hex_to_bin (void) {
	/* To begin, assume the lowest address is at the end of the memory.
	subsequent addresses will lower this number. At the end of the input
	file, this value will be the lowest address. */
	Segment = 0;
	Upper_Address = 0;
	Lowest_Address = MEMORY_SIZE - 1;
	Highest_Address = 0;

	/* Now read the file & process the lines. */
	do /* repeat until EOF(Filin) */
	{
		fgets(Line,MAX_LINE_SIZE,Filin);  		          /* Read a line from input file. */
		i = strlen(Line)-1;                          		/* Remove carriage return/line feed at the end of line. */
		if (Line[i] == '\n') Line[i] = '\0';
		/* Scan the first two bytes and nb of bytes.
		The two bytes are read in First_Word since it's use depend on the
		record type: if it's an extended address record or a data record.
		*/
		sscanf (Line, ":%2x%4x%2x%s",&Nb_Bytes,&First_Word,&Type,Data_Str);
		Checksum = Nb_Bytes + (First_Word >> 8) + (First_Word & 0xFF) + Type;
		p = Data_Str;
		/* If we're reading the last record, ignore it. */
		switch (Type) {
			/* Data record */
			case 0:
				Address = First_Word;
				if (Seg_Lin_Select == SEGMENTED_ADDRESS)
					Phys_Addr = ((Segment << 4) & ADDRESS_MASK) + Address;
				else
					/* LINEAR_ADDRESS or NO_ADDRESS_TYPE_SELECTED
					Upper_Address = 0 as specified in the Intel spec. until an extended address
					record is read. */
					Phys_Addr = ((Upper_Address << 16) & ADDRESS_MASK) + Address;

				/* Check that the physical address stays in the buffer's range. */
				if ((Phys_Addr + Nb_Bytes) <= MEMORY_SIZE -1) {
					/* Set the lowest address as base pointer. */
					if (Phys_Addr < Lowest_Address)
						Lowest_Address = Phys_Addr;

					/* Same for the top address. */
					temp = Phys_Addr + Nb_Bytes -1;

					if (temp > Highest_Address)
						Highest_Address = temp;

					/* Read the Data bytes. */
					/* Bytes are written in the Memory block even if checksum is wrong. */
					for (i= Nb_Bytes; i > 0; i--) {
						sscanf (p, "%2x",&temp2);
						p += 2;
						Memory_Block[Phys_Addr++] = temp2;
						Checksum = (Checksum + temp2) & 0xFF;
					};

					/* Read the Checksum value. */
					sscanf (p, "%2x",&temp2);

					/* Verify Checksum value. */
					Checksum = (Checksum + temp2) & 0xFF;

					if ((Checksum != 0) && Enable_Checksum_Error) {
						Status_Checksum_Error = TRUE;
					}
				} else {
					if (Seg_Lin_Select == SEGMENTED_ADDRESS)
						fprintf(stderr,"Data record skipped at %4X:%4X\n",Segment,Address);
					else
						fprintf(stderr,"Data record skipped at %8X\n",Phys_Addr);
				}

				break;

				/* End of file record */
			case 1:
				/* Simply ignore checksum errors in this line. */
				break;

				/* Extended segment address record */
			case 2:
				/* First_Word contains the offset. It's supposed to be 0000 so
				we ignore it. */

				/* First extended segment address record ? */
				if (Seg_Lin_Select == NO_ADDRESS_TYPE_SELECTED)
					Seg_Lin_Select = SEGMENTED_ADDRESS;

				/* Then ignore subsequent extended linear address records */
				if (Seg_Lin_Select == SEGMENTED_ADDRESS) {
					sscanf (p, "%4x%2x",&Segment,&temp2);

					/* Update the current address. */
					Phys_Addr = (Segment << 4) & ADDRESS_MASK;

					/* Verify Checksum value. */
					Checksum = (Checksum + (Segment >> 8) + (Segment & 0xFF) + temp2) & 0xFF;

					if ((Checksum != 0) && Enable_Checksum_Error)
						Status_Checksum_Error = TRUE;
				}
				break;

				/* Start segment address record */
			case 3:
				/* Nothing to be done since it's for specifying the starting address for
				execution of the binary code */
				break;

				/* Extended linear address record */
			case 4:
				/* First_Word contains the offset. It's supposed to be 0000 so
				we ignore it. */

				/* First extended linear address record ? */
				if (Seg_Lin_Select == NO_ADDRESS_TYPE_SELECTED)
					Seg_Lin_Select = LINEAR_ADDRESS;

				/* Then ignore subsequent extended segment address records */
				if (Seg_Lin_Select == LINEAR_ADDRESS) {
					sscanf (p, "%4x%2x",&Upper_Address,&temp2);

					/* Update the current address. */
					Phys_Addr = (Upper_Address << 16) & ADDRESS_MASK;

					/* Verify Checksum value. */
					Checksum = (Checksum + (Upper_Address >> 8) + (Upper_Address & 0xFF) + temp2)
							& 0xFF;

					if ((Checksum != 0) && Enable_Checksum_Error)
						Status_Checksum_Error = TRUE;
				}
				break;

				/* Start linear address record */
			case 5:
				/* Nothing to be done since it's for specifying the starting address for
				execution of the binary code */
				break;
			default:
				break;
		}
	} while (!feof (Filin));
	/*-----------------------------------------------------------------------------*/

	fprintf(stdout,"HEXCONVERT: Lowest address = %08X\n",Lowest_Address);
	fprintf(stdout,"HEXCONVERT: Highest address = %08X\n",Highest_Address);

	/* This starting address is for the binary file,

	ex.: if the first record is :nn010000ddddd...
	the data supposed to be stored at 0100 will start at 0000 in the binary file.

	Specifying this starting address will put FF bytes in the binary file so that
	the data supposed to be stored at 0100 will start at the same address in the
	binary file.
	*/

	if(Starting_Address_Setted) {
		Lowest_Address = Starting_Address;
	}
	//	free (Memory_Block);

#ifdef USE_FILE_BUFFERS
	free (FilinBuf);
#endif

	fclose (Filin);

	if (Status_Checksum_Error & Enable_Checksum_Error) {
		fprintf(stderr,"Checksum error detected.\n");
		return 1;
	}

	return 0;
}
