/***************************************************************************
 *   Copyright (C) 2005 by Pavel Vymetálek   *
 *   pvymetalek@seznam.cz   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   --------------------------------------------------------------------  *
 *   SIMEPROM - control program for Elnec's SIMEPROM-01B                   *
 *   Device manufacturer www.elnec.sk                                      *
 *   Linux control program for Linux and other unices by Pavel Vymetálek          *
 ***************************************************************************/

/* ------------------------------------------------------------------------
   Low Level code for Linux PIC programmer
   
   Copyright 1994-2002 by Brian C. Lane
   All Rights Reserved
 
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.
 
   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 
   ---------------------------------------------------------------------
   This is the way Russ has his programmer connected
   to the parallel port.
 
   DB25 connections
    1 = !C0    = !PGM
    2 = D0     = CLK out
    3 = D1     = DATI (to PIC)
    11 = !S7   = DATO (from PIC)
    14 = !C1   = MODE control 0=programming  1=normal
    25 = GND
 
   David Tait's Parallel port PIC programmer connections:
   May or may not be inverted!
      
   2 = D0      = Data out
   3 = D1      = Clock out
   4 = D2      = +5v Power control
   5 = D3      = +Vpp/!MCLR control
   10= !ACK    = Data in
 
   ------------------------------------------------------------------------
   08/23/2002   Not sure what was wrong, but it works fine today. It also
   bcl          appears that the pconfig.delay setting isn't really needed
                for anything. The waitus() is sufficient for the timing
		needed. Removing it, but leaving the waitus(1) calls in
		inb and outb as suggested by Lee Olsen.
 
   08/22/2002   Changes to biton and off where ~ was changed to ! was wrong
   bcl          It now reads the PIC. It even erases it to 0x3FFF but when
                I try to program it, all attempts return verify errors.
		Changed all waitus() calls to use pconfig.delay to control
		its delay time. Still won't program.
 
   08/21/2002   Changed the waitus calls in read_pic() to use pconfig.delay
   bcl          to see if slowing things down will help it work correctly.
 
 
   08/19/2002   Changed ioperm to start at port, not port+1
   bcl
 
   08/18/2002   Changing how I address the port. The kernel support for the
   bcl          direct I/O interface has changed and it split up into parts
                that only appear in the kernel source. The only real piece
		of information needed is the base address. Removing the
		lp_struct and LP_B usage. Using lp_table as a list of port
		addresses.
		Changing the names of inb and outb to pp_ so that they do
		not conflict with already existing kernel functions.
 
		Changing inb and outb to match the current /sys/io.h defines
		I still get warnings about them though.
 
   08/16/2002   Adding changes from Lee Olsen to read /proc/ioports to get
   bcl          port information and support PCI PIO cards that have high
                io port addresses (0xC400 in his case). Reportedly also
		works on 2.2.x kernels.
		Use iopl(3) instead of ioperm() for large addresses.
		Added waitus(1) calls in inb and outb to make things work,
		and changed ~inb to !inb in inbit()
		Q. about this last change -- what's the diff between ! and ~?
 
		I don't like using iopl, it isn't specific enough, giving
		access to all ports as far as I can tell. A really bad thing!
 
   07/05/98 bcl Adding TRUE/FALSE return from verify_pic
                Adding quiet mode (from options & OPT_QUIET)
 
   11/15/97 bcl Revising this software to work with kernel v2.0.31
   		I seem to have lost my code that lets the user specify
   		the parallel port pin number and state for each control
   		line. I guess I'll have to rewrite that once I get this
   		functioning in its current state. Having trouble with
   		ioperm... Fixed, needed a 1L passed, not a 1. I really
   		hate that.
   		Adding pin definitions.
                Chaninging everything so that it uses the pin_table.
                Pretty cool!
 
   11/22/94 bcl Thanksgiving Vacation. Should have the time to finish this.
 
   11/21/94 bcl	I changed over to using the function pointer array. Made
   		everything work 'Really Badly'!
 
   ------------------------------------------------------------------------ */

/** @TODO Volba portu "LPT0,1,2,3,4,5,..." jako spousteci parametr a jeste lepe definovany jako /dev/lp0 .. apod
 * @TODO zkusit nedelat to prilis nizkourovnove... tj. nepouzivat assembler...
*/

#include <unistd.h>
#include <sys/io.h>
#include <fcntl.h>
#include <linux/lp.h>
#include <linux/ppdev.h>
#include <sys/ioctl.h>
#include <sys/time.h>
//#include <asm/system.h>
#include <linux/parport.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include <simeprom7.h>
#include <lowlevel.h>


#define NUM_PORTS 10
#define OPT_DEBUG       0x0378		// puvodne 0004
#define TRUE	1
#define FALSE	0

int cc;	// adresa v eeprom


/* How to genericly handle the ports, lines, and polarities? */
struct _pconfig {
	int	port;				/* /dev/lpX		 	*/
	int	colors;				/* Color support 1 or 0		*/

	int	VPP_pin;			/* Pin for VPP control		*/
	int	VPP_pol;			/* Polarity 0=inverted 1=normal	*/
	int	VDD_pin;			/* Pin for Vdd control		*/
	int	VDD_pol;			/* Polarity 0=inverted 1=normal	*/
	int	CLK_pin;			/* Pin for Clock control	*/
	int	CLK_pol;			/* Polarity 0=inverted 1=normal	*/
	int	DATAO_pin;			/* Pin for Data to PIC control	*/
	int	DATAO_pol;			/* Polarity 0=inverted 1=normal	*/
	int	DATAI_pin;			/* Pin for Data from PIC control*/
	int	DATAI_pol;			/* Polarity 0=inverted 1=normal	*/
};


struct _pconfig		pconfig; 		/* port configuration		*/


extern int options;			/* Options 			*/
static unsigned lp_data[3];		/* Current levels		*/
int port_fd = -1;
int max_lp;
int	        options=0;
unsigned char buff[0x0ffff];				/* buffer 64k */
int port;
char last_value;




/*extern struct _pic pic;*/
/*extern struct _pconfig pconfig;*/

/* Table of parallel port addresses, maximum of 6 supported */
unsigned int lp_table[NUM_PORTS];

/*
   Base + 0   Data I/O Address    TTL Output Only/(Tristate->Input)
   ---------------------------------------------------------------------
   Bit:  7       6       5       4       3       2       1       0
   Pin:  9       8       7       6       5       4       3       2
   ---------------------------------------------------------------------
 
   Base + 1   Status I/O Address  TTL Input Only
   ---------------------------------------------------------------------
   Bit:  7 Inv   6       5       4       3       2       1       0
   Pin:  11      10      12      13      15      NC      NC      NC
   ---------------------------------------------------------------------
 
   Base + 2   Control I/O Address Open-Collector Output/(Input:WR=>04h)
   ---------------------------------------------------------------------
   Bit:  7       6       5       4       3 Inv   2       1 Inv   0 Inv
   Pin:  NC      NC  NC/(BiDir) Int      17      16      14      1 
   ---------------------------------------------------------------------
 
   Copyright to Public Domain by Richard Steven Walz (C) 1996
   ftp://ftp.armory.com/pub/user/rstevew/LPT/simplpt.faq
 
   This defines the pin # to ioport mappings. offset is relative to
   lp_table[ pconfig.port ] and mask is the pins bit position in the byte
 */
struct {
	unsigned int	offset;
	unsigned char	mask;
} pin_table[26] = {
                    { 1, 0x00 },			/* pin 0	Invalid pin #	*/
                    { 2, 0x01 },			/* pin 1	(out)	!strobe	*/
                    { 0, 0x01 },			/* pin 2	(out)	Data 0	*/
                    { 0, 0x02 },			/* pin 3	(out)	Data 1	*/
                    { 0, 0x04 },			/* pin 4	(out)	Data 2	*/
                    { 0, 0x08 },			/* pin 5	(out)	Data 3	*/
                    { 0, 0x10 },			/* pin 6	(out)	Data 4	*/
                    { 0, 0x20 },			/* pin 7	(out)	Data 5	*/
                    { 0, 0x40 },			/* pin 8	(out)	Data 6	*/
                    { 0, 0x80 },			/* pin 9	(out)	Data 7	*/
                    { 1, 0x40 },			/* pin 10	(in)	!ack	*/
                    { 1, 0x80 },			/* pin 11	(in)	busy	*/
                    { 1, 0x20 },			/* pin 12	(in)	Pout	*/
                    { 1, 0x10 },			/* pin 13	(in)	Select	*/
                    { 2, 0x02 },			/* pin 14	(out)	!feed	*/
                    { 1, 0x08 },			/* pin 15	(in)	!error	*/
                    { 2, 0x04 },			/* pin 16	(out)	!init	*/
                    { 2, 0x08 },			/* pin 17	(out)	!SI	*/
                    { 1, 0x00 },			/* pin 18	GND		*/
                    { 1, 0x00 },			/* pin 19	GND		*/
                    { 1, 0x00 },			/* pin 20	GND		*/
                    { 1, 0x00 },			/* pin 21	GND		*/
                    { 1, 0x00 },			/* pin 22	GND		*/
                    { 1, 0x00 },			/* pin 23	GND		*/
                    { 1, 0x00 },			/* pin 24	GND		*/
                    { 1, 0x00 },			/* pin 25	GND		*/
                };


/* ------------------------------------------------------------------------
   Delay for a number of uS -- Works better than usleep for me!
   ------------------------------------------------------------------------ */
void waitus( unsigned int us ) {
	struct timeval tv, wt;

	gettimeofday( &tv, NULL );

	wt.tv_usec = ( tv.tv_usec + us ) % 1000000;
	wt.tv_sec = tv.tv_sec + ( ( tv.tv_usec + us ) / 1000000);

	/* loop until time is up */
	for (;;) {
		gettimeofday( &tv, NULL );
		if ( tv.tv_sec > wt.tv_sec )
			return;
		if ( tv.tv_sec == wt.tv_sec )
			if ( tv.tv_usec > wt.tv_usec )
				return;
	}
}



/* -----------------------------------------------------------------------
   Port control routines -- easier than asm/io.h's routines
   ----------------------------------------------------------------------- */


static inline void pp_outb ( char value, short port) {
	value &= 0x0ff;
	//  if (DEBUG == 1 && cc < 0x30) printf("port = 0x%Xh\n", value);

	//  waitus( 1 );

__asm__ __volatile__ ("outb %b0,%w1": :"a" (value), "Nd" (port));
	last_value = value;
}

/*
static inline unsigned char pp_inb (short port) {
	unsigned int ret;

	waitus( 1 );

__asm__ __volatile__ ("inb %w1,%0":"=a" (ret):"Nd" (port));

	return ret;
}
*/


/* -----------------------------------------------------------------------
	Low Level programmer control
	pconfig has the pins and polarity of the various control lines
	pin_table[] has the offset and mask for eack of the DB-25 pins.
	--------------------------------------------------------------------- */

/* turn on an output */
/*
int biton( int pin, int pol ) {
	if( pol == 1 ) {
		lp_data[pin_table[pin].offset] |= pin_table[pin].mask;
	} else {
		lp_data[pin_table[pin].offset] &= ~pin_table[pin].mask;
	}
	pp_outb( lp_data[pin_table[pin].offset], lp_table[pconfig.port] + pin_table[pin].offset );

	return TRUE;
}
*/
/* turn off an output */
/*
int bitoff( int pin, int pol ) {
	if( pol == 1 ) {
		lp_data[pin_table[pin].offset] &= ~pin_table[pin].mask;
	} else {
		lp_data[pin_table[pin].offset] |= pin_table[pin].mask;
	}
	pp_outb( lp_data[pin_table[pin].offset], lp_table[pconfig.port] + pin_table[pin].offset );

	return TRUE;
}
*/


/* Input a bit */
/*
int inbit() {
	// If the data is not inverted, return the normal state. Otherwise,
	//				invert it
	if( pconfig.DATAI_pol == 1 ) {
		return( pp_inb( lp_table[pconfig.port] + pin_table[pconfig.DATAI_pin].offset )
		        & pin_table[pconfig.DATAI_pin].mask );
	} else {
		return( !pp_inb( lp_table[pconfig.port] + pin_table[pconfig.DATAI_pin].offset )
		        & pin_table[pconfig.DATAI_pin].mask );
	}
}

*/


/* Clock high */
//int clkhi()
//{
//  return( biton( Pin2, 1 ) );
//}

/* Clock low */
//int clklo()
//{
//  return( bitoff( pconfig.CLK_pin, pconfig.CLK_pol ) );
//}

/* Data high */
//int outhi()
//{
//  return( biton( pconfig.DATAO_pin, pconfig.DATAO_pol ) );
//}

/* Data low */
//int outlo()
//{
//  return( bitoff( pconfig.DATAO_pin, pconfig.DATAO_pol ) );
//}

/* -----------------------------------------------------------------------
				Find the port information from /proc/ioports
 
				Lee just used the current table (3 entries).
				It ought to be expanded to dynamically allocate space for any number of
				parallal ports.
				----------------------------------------------------------------------- */
void init_lp_table( void ) {
	char buf[80];
	FILE *fp;
	int  start;
	int  stop;
	int  id;

	for( id=0; id<NUM_PORTS; id++ ) {
		lp_table[id] = 0;
	}
	max_lp = 0;

	/* Read /proc/ioports and get the base address for parport entries */
	if( (fp = fopen( "/proc/ioports", "r" )) != NULL ) {
		while( fgets( buf, 80, fp ) != NULL ) {
			/* search for parallel ports and add them to the list */
			if( sscanf( buf, "%x-%x : parport%d", &start, &stop, &id ) == 3 ) {
				if( id < NUM_PORTS ) {
					lp_table[id] = start;
					if(id > max_lp) {
						max_lp = id;
					}
				} /* NUM_PORTS */
			} /* sscanf */
		} /* readin /proc/ioports */
	} /* opening /proc/ioports */
	fclose( fp );
}



/* -----------------------------------------------------------------------
				Try to initalize the LP port, find it's base address etc. to make sure
				that it is useable.
				----------------------------------------------------------------------- */
int init_port() {
	char	printer[10];
	int	fd;

	/* Make sure the port number is within the proper range */
	if( (pconfig.port > NUM_PORTS) || (pconfig.port > max_lp) ) {
		fprintf(stderr, "Error, lp%d is not supported, %d is > maximum port %d\n",
		        pconfig.port, pconfig.port, max_lp );
		return 0;
	}

	if( lp_table[pconfig.port] == 0 ) {
		fprintf(stderr, "Error, base address for %d is 0x0000\n", pconfig.port );
		return 0;
	}

	//					sprintf(printer,"/dev/lp%d", pconfig.port );
	sprintf(printer,"/dev/parport%d", pconfig.port );

	/* check for access to the printer */
	if( access( printer, R_OK|W_OK ) < 0 ) {
		fprintf( stderr, "You don't have permission to access %s\n", printer );
		return 0;
	}

	fd=open(printer,O_WRONLY);
	if ( fd<0 ) {
		perror(NULL);
		return 0;
	}

	if ( options & OPT_DEBUG )
		printf( "I/O base address for %s is 0x%x\n", printer, lp_table[pconfig.port] );

	if( lp_table[pconfig.port] < 0x3ff ) {
		/* Get permission to access the config register */
		if ( ioperm( lp_table[pconfig.port], 3L, 1) <0 ) {
			fprintf( stderr, "Access to port 0x%x denied\n", lp_table[pconfig.port] );
			return 0;
		}
	} else {
		/* High port address, try to get access. This is a really, really bad
						way to do this, as it exposes ALL the ports I believe.
		*/
		if( iopl(3) < 0 ) {
			perror(NULL);
			fprintf( stderr, "Access to port 0x%x denied\n", lp_table[pconfig.port]+1 );
			return 0;
		}
	}

	/* Get the initial states of data and command line registers */
	/* zakomentovano v souvislosti s experimenty pres ioctl
	lp_data[0] = pp_inb( lp_table[pconfig.port] + 0 );
	lp_data[1] = pp_inb( lp_table[pconfig.port] + 1 );
	lp_data[2] = pp_inb( lp_table[pconfig.port] + 2 );
	*/
	
	//  if ( options & OPT_DEBUG )
	//    printf( "data = %02X   status = %02X   cmnd = %02X\n",
	//	    lp_data[0], lp_data[1], lp_data[2] );
	return 1;
}

/*********************************************************************************/
/*********************************************************************************/
/*********************************************************************************/
/*********************************************************************************/
/*********************** S  I  M  E  P  R  O  M **********************************/

/* nastavenie paralelneho portu do vychodiskoveho stavu*/
void silence(void) {
	unsigned char pst;
	pst = 0;
	pst = resethi | strhi | clkhi | datahi;
	//port = 0x0378;
	//pp_outb ( pst, port);
	ioctl (fd, PPWDATA, &pst);
}

/* aktivny RESET */
void rst (void) {
	unsigned char pom;
//	port = 0x0378;
//	pp_outb (last_value & resetlo, port);
	pom = last_value & resetlo;
	ioctl (fd, PPWDATA, &pom);
	printf ("Reset ON\n");
}

/* RESET neaktivny */
void norst (void) {
	unsigned char pom;
	//port = 0x0378;
	//pp_outb (last_value | resethi, port);
	pom = last_value | resethi;
	ioctl (fd, PPWDATA, &pom);	
	printf ("Reset OFF\n");
}

/* nastavenie simulatora do modu DOWNLOAD */
void load (void) {
	/* puvodni kod
	port = 0x0378;
	pp_outb (loadhi, port+2);
	*/
	// nove
	struct ppdev_frob_struct frob;
	frob.mask = PARPORT_CONTROL_INIT;
	frob.val = 0; //PARPORT_CONTROL_INIT;
	ioctl (fd, PPFCONTROL, &frob);
	//printf ("shozen init drat...\n");
	//sleep(3);
}

/* nastavenie simulatora do modu simulacia */
void sim (void) {
	/*
	port = 0x0378;
	pp_outb (loadlo, port+2);
	*/
	/*
	// port = 0x0378;
	// pp_outb (resethi, port);
	*/
	struct ppdev_frob_struct frob;
	frob.mask = PARPORT_CONTROL_INIT;
	frob.val = PARPORT_CONTROL_INIT;
	ioctl (fd, PPFCONTROL, &frob);
	//printf ("shozen init drat...\n");
	//sleep(3);

}

/* download jedneho byte dat do simulatora. Parametre procedury su adresa byte
				dat a vlastne data, adresa paralelneho portu je globalna premenna alebo
				konstanta */
void put_byte (unsigned int padr, unsigned char pbyte) {
	int ahi, alo;					/* adresa rozdelena na HIGH a LOW */
	int ob;						/* pomocna premenna pre tvorbu byte dat na LPT port */
	int oldb;						/* odpamatana hodnota na porte LPT */
	int cn;						/* pocitadlo */
	int mask;						/* putujuca jednotka, pomocou ktorej sa vybera z daneho BYTE adresy a dat */
	
	unsigned int pom;

	// oldb = last_value & 0x0F8 & clklo;		/* and F8 - vynulovanie bitov DATA, ADRLO a ADRHI, bit CLK do 0 */
	oldb = strhi;		/* and F8 - vynulovanie bitov DATA, ADRLO a ADRHI, bit CLK do 0 */
	ob = oldb;
	mask = 1;						/* download ide od najnizsieho bitu daneho byte */
	ahi = padr & 0x0FF00;
	ahi = ahi >> 8;
	alo = padr & 0x0FF;

	alo = 0x0ff - alo;
	ahi = 0x0ff - ahi;

	for (cn = 0; cn <= 7; cn++) {
		ob = oldb;					/* najpr zapis pociatocnej hodnoty na port */
		if ((ahi & mask) != 0) ob = ob | ahihi;
		if ((alo & mask) != 0) ob = ob | alohi;
		if ((pbyte & mask) != 0) ob = ob | datahi;
		port = 0x0378;
		//pp_outb (ob, port);			/* potom zapis spravnych bitov byte dat a adresy */
		pom = ob;
		ioctl (fd, PPWDATA, &pom);
		
		mask = mask << 1;
		port = 0x0378;
		
		//pp_outb (ob | clkhi, port);
		pom = ob | clkhi;
		ioctl (fd, PPWDATA, &pom);
		
		waitus(1);
		//pp_outb (ob, port);		/* potvrdit platnost hranou CLK, teda  __/~\__ */
		pom = ob;
		ioctl (fd, PPWDATA, &pom);
		
	}
	port = 0x0378;
	//pp_outb(ob & strlo, port);
	pom = ob & strlo;
	ioctl (fd, PPWDATA, &pom);
 
	waitus(1); 
	//pp_outb (ob, port);			/* strobe ~\_/~ ; dany byte dat zapisat na danu adresu internej CMOS RAM pulzom do L */
	pom = ob;
	ioctl (fd, PPWDATA, &pom);

}

void program () {
	// testovani
	 
	silence();
	load();
	rst();
	printf ("Start downloading... \n");
	for (cc = 0; cc <= 0x0ffff; cc++) {
		if ((cc & 0x0f) == 0) printf ("addr: %X  \r", cc);
		put_byte (cc, Memory_Block[cc]);
	}
	printf ("addr: %X  \r", cc-1);
	printf ("\nDownloading finished...\n");
	sim();
	norst();
}
