/***************************************************************************
 *   Copyright (C) 2005 by Pavel Vymetálek   *
 *   pvymetalek@seznam.cz   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   --------------------------------------------------------------------  *
 *   SIMEPROM - control program for Elnec's SIMEPROM-01B                   *
 *   Device manufacturer www.elnec.sk                                      *
 *   Linux control program for Linux and other unices by Pavel Vymetálek   *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <simeprom7.h>
#include <hex2bin.h>
#include <lowlevel.h>

#include <unistd.h>
#include <sys/io.h>
#include <fcntl.h>
#include <linux/lp.h>
#include <sys/ioctl.h>
#include <sys/time.h>
//#include <asm/system.h>
#include <linux/ppdev.h>
#include <ieee1284.h>

//extern int max_lp;
//extern void program(void);
//extern int init_port(void);
//extern void init_lp_table(void);
//extern int readihex( FILE *fp, unsigned char *mem);


void *NoFailMalloc (size_t size);
void NoFailOpenInputFile (char *Flnm);
void NoFailOpenOutputFile (char *Flnm);

/* flag that a file was read */
boolean Fileread;
boolean Enable_Checksum_Error = FALSE;
boolean Status_Checksum_Error = FALSE;

/* We will assume that when one type of addressing is selected, it will be valid for all the
current file. Records for the other type will be ignored. */
unsigned int	Seg_Lin_Select = NO_ADDRESS_TYPE_SELECTED;

boolean Starting_Address_Setted = FALSE;

char parport_device[100] = "/dev/parport0";
int fd;		// file device - desctriptor of paralel port


/***********************************************************************/

void *NoFailMalloc (size_t size) {
	void *result;

	if ((result = malloc (size)) == NULL) {
		fprintf (stderr,"Can't allocate memory.\n");
		exit(1);
	}
	return (result);
}

/* Open the input file, with error checking */
void NoFailOpenInputFile (char *Flnm) {
	while ((Filin = fopen(Flnm,"r")) == NULL) {
		fprintf (stderr,"Input file %s cannot be opened. Enter new filename: ",Flnm);
		fgets (Flnm,MAX_FILE_NAME_SIZE,stdin);
		if (Flnm[0] == '\0') exit(1);
	}

#ifdef USE_FILE_BUFFERS
	FilinBuf = (char *) NoFailMalloc (BUFFSZ);
	setvbuf(Filin, FilinBuf, _IOFBF, BUFFSZ);
#endif
} /* procedure OPENFILIN */





/********************************************************************/
/********************************************************************/

void usage(void) {
	fprintf (stderr,
		   "\nusage: simeprom [-s xxxx] [-c] filename\n"
				   "-s xxxx\tStarting address in hex, default is 0x0000\n"
				   "-d parport device\texample: /dev/parport0, default is /dev/parport0\n"
				   "-c enable checksum verification, default is no verification\n\n");

	exit(1);
} /* procedure USAGE */


void InitApp(void) {
	fprintf (stdout,"Simeprom v0.7.0, Copyright (C) 2003 - 2005 Pavel Vymetalek, (C) 1999 Jacques Pelletier (hexbin utility)\n");
}

void TestArgs(int argc, char *argv[]) {
	if (argc == 1) {
		usage();
	}
	Param = 1;
	for (Param = 1; Param < argc; Param++) {
		c = p = argv[Param];
		if(*c == '-') {
			p = c + 2;
			/* Parameter may follow immediately after option */
			/* Skips after parameter to next option */
			if (*p == '\0') p = argv[++Param];
			switch(tolower (*(++c))) {
				case 'c':
					Enable_Checksum_Error = TRUE;
					break;
				case 's':
					sscanf(p,"%x",&Starting_Address);
					Starting_Address_Setted = TRUE;
					break;
				case 'd':
					sscanf(p,"%s",&parport_device);
					break;
				case 'h':
					usage();
					break;
				default:
					usage();
			} /* switch */
		} else
			break;    /* start to get next param */
			/* if option */
	} /* for Param */
}

void LoadHexFile(int argc, char *argv[]) {
	/* when user enters input file name */
	/* Assume last parameter is filename */
	strcpy(Filename,argv[argc -1]);

	/* Just a normal file name */
	NoFailOpenInputFile (Filename);
	Fileread = TRUE;
}

void HexToBin(void) {
	Memory_Block = (byte *) NoFailMalloc(MEMORY_SIZE);  	/* allocate a buffer */
	memset (Memory_Block,0xFF,MEMORY_SIZE);             	/* For EPROM or FLASH memory types, fill unused bytes with FF */
										/* Memory_Block contains no eprom data, but it's initialized */
	hex_to_bin();
}

int UploadEprom(void) {
	program();
	Memory_Block[0] = 0;
	free (Memory_Block);
	return 0;                 /* success */
}
/********************************************************************/

//------------------------------------------------------------------------------
 int OpenParalelPort(const char *name) {
	 int mode; /* We'll need this later. */
	 printf ("Pokus o otevreni paralelniho portu: %s... ", name);
	 fd = open (name, O_RDWR);
	 //printf ("fd: %d\n", fd);
	 printf ("OK\n");
	 if (fd == -1) {
		 perror ("can not open paralel port");
		 return 1;
	 }
	 if (ioctl (fd, PPCLAIM)) {
		 perror ("PPCLAIM");
		 close (fd);
		 return 1;
	 }
   /* Switch to compatibility mode.  (In fact we don't need
    * to do this, since we start off in compatibility mode
    * anyway, but this demonstrates PPNEGOT.) 
    */
	 //mode = IEEE1284_MODE_COMPAT;
	 mode = M1284_COMPAT;
	 if (ioctl (fd, PPNEGOT, &mode)) {
	 perror ("PPNEGOT");
	 close (fd);
	 return 1;
 }


	 return 0;

 }


//------------------------------------------------------------------------------
int main(int argc, char *argv[]) {
	InitApp();
	TestArgs(argc, argv);
	LoadHexFile(argc, argv);
	
	if (OpenParalelPort(parport_device)) {
		printf("nemohu ponovu otevrit port\n");
		 exit (1); // test otecreni portu po novu
	}
	
	
	if (!Fileread) {
		usage();
		exit(1);
	}
	HexToBin();
	UploadEprom();
	return EXIT_SUCCESS;
}
//------------------------------------------------------------------------------
